package soal.markov

import java.util.Random

import com.twitter.cassovary.graph.{DirectedGraph, Node}
import soal.util.{CollectionsUtil, DiscreteAliasSampler}

import scala.collection.mutable

/**
 * Contains methods related to personalized PageRank estimation.  All methods operate in the
 * context of the given graph, teleportProbability, and random.
 */

class BidirectionalTransitionEstimator (val graph: DirectedGraph[Node],
                                        val heatKernelT: Double,
                                        val maxLength: Int,
                                        val random: Random = new Random) {


  def heatKernel(sourceId: Int,
                         targetId: Int,
                         minimumProbability: Double,
                         relativeError: Double,
                         chernoffConstantOption: Option[Double]): Double = {
    val provableChernoffConstant = 6 * math.exp(1) * math.log(2 * maxLength / 1.0e-6) // guarantees the given relative error with
    // probability 1 - 1.0e-6
    val chernoffConstant = chernoffConstantOption.getOrElse(provableChernoffConstant)

    val chernoffMultiplier = chernoffConstant / math.pow(relativeError, 2.0).toDouble
    // TODO: Add balance constant and tune in experiments, or implement balanced version
    val reverseThreshold = math.sqrt(minimumProbability / chernoffMultiplier).toDouble
    val walkCount = (chernoffMultiplier * reverseThreshold / minimumProbability).toInt
    val (estimates, residuals) = computeContributions(targetId, reverseThreshold)
    val estimate =
      ((0 until walkCount) map { walkIndex =>
        val length = sampleHeatKernelLength()
        val vs = samplePath(sourceId, length + 1)
        estimates(length)(sourceId) + ((0 to length) map { i =>
          residuals(length - i)(vs(i))
        }).sum
      }).sum / walkCount
    estimate
  }


  /** Given a target nodeId, returns two Sequencies of maps (estimates,
    * residuals). The value estimates(k)(t)
    * approximates Pr[s -> t in k steps] with additive error maxResidual,
    * and the residuals satisfy a useful loop invariant.  See the paper for details.
    */
  def computeContributions(targetId: Int, maxResidual: Double
      ): (IndexedSeq[mutable.Map[Int, Double]], IndexedSeq[mutable.Map[Int, Double]]) = {

    // estimates(k)(uId) estimates P^k(u, target)
    // Use maxLength + 2 so we can push back maxLength steps from target
    val estimates = Array.fill(maxLength + 2)(CollectionsUtil.efficientIntDoubleMap()
      .withDefaultValue(0.0f))
    val residuals = Array.fill(maxLength + 2)(CollectionsUtil.efficientIntDoubleMap()
      .withDefaultValue(0.0f))
    residuals(0)(targetId) =  1.0f

    // Store all nodes with residual greater than maxResidual in a queue to be processed
    // We will processes the residuals in layers: First k=0 (where k is distance back from target)
    // then k=1, k=2, ...
    // Use ArrayDequeue because it is more efficient than the linked-list based mutable.Queue
    var largeResidualNodes = new java.util.ArrayDeque[Int]()
    largeResidualNodes.add(targetId)
    for (k <- 0 to maxLength) {
      val nextLargeResidualNodes = new java.util.ArrayDeque[Int]()
      while (!largeResidualNodes.isEmpty) {
        val vId = largeResidualNodes.pollFirst()
        val kvResidual = residuals(k)(vId)
        estimates(k)(vId) += kvResidual
        residuals(k)(vId) = 0.0f
        val v = graph.getNodeById(vId).get
        for (uId <- v.inboundNodes()) {
          val u = graph.getNodeById(uId).get
          val residualChange = 1.0f / u.outboundCount * kvResidual
          if (k + 1 < maxLength &&
            residuals(k+1)(uId) < maxResidual &&
            residuals(k+1)(uId) + residualChange >= maxResidual)
            nextLargeResidualNodes.add(uId)
          residuals(k+1)(uId) += residualChange
        }
      }
      largeResidualNodes = nextLargeResidualNodes
    }

    (estimates, residuals)
  }

  /**
   * Samples from the PageRank distribution personalized to the given source distribution by
   * doing a single random walk.
   */
  def samplePath(sourceId: Int, length: Int): collection.IndexedSeq[Int] = {
    var v = sourceId
    val path = Array.fill(length)(-1) // If we reach dead end, jump to sink state "-1"
    for (i <- 0 until length) {
      path(i) = v
      val vNode = graph.getNodeById(v).get
      vNode.randomOutboundNode(random) match {
        case Some(outNeighbor) => v = outNeighbor
        case None => return path
      }
    }
    path
  }

  // For sampling from heat kernel
  def heatKernelProbability(i: Int): Float =
    (math.exp(-heatKernelT) * math.pow(heatKernelT, i) / factorial(i)).toFloat
  val unnormalizedLengthProbabilities = (0 to maxLength) map heatKernelProbability

  val lengthSampler = new DiscreteAliasSampler(
    0 to maxLength, unnormalizedLengthProbabilities, random
  )
  def sampleHeatKernelLength(): Int = {
    lengthSampler.sample()
  }

  def factorial(k: Int): Double = {
    if (k >= 2)
      ((1 to k) map {
        _.toDouble
      }).product
    else
      1
  }

  def computePhis(): IndexedSeq[Double] = {
    val phis = new Array[Double](maxLength + 1)
    phis(maxLength) = 1
    for (i <- maxLength - 1 to 0 by -1) {
      phis(i) = phis(i + 1) * heatKernelT / (i + 1.0) + 1
    }
    phis
  }
  val phis = computePhis()
  // Based on David Gleich's paper
  def forwardPushHeatKernel(sourceId: Int, epsilon: Double): collection.Map[Int, Double] = {

    // estimates(k)(uId) estimates P^k(u, target)
    // Use maxLength + 2 so we can push back maxLength steps from target
    val estimates = CollectionsUtil.efficientIntDoubleMapWithDefault0()
    val residuals = Array.fill(maxLength + 2)(CollectionsUtil.efficientIntDoubleMapWithDefault0())
    residuals(0)(sourceId) =  1.0 / math.exp(heatKernelT)

    // Store all (node, k == forwardDistance) with residual greater than maxResidual in a queue to
    // be processed

    // Use ArrayDequeue because it is more efficient than the linked-list based mutable.Queue
    var largeResidualNodes = new java.util.ArrayDeque[(Int, Int)]()
    largeResidualNodes.add((sourceId, 0))
    for (k <- 0 to maxLength) {
      while (!largeResidualNodes.isEmpty) {
        val (vId, k) = largeResidualNodes.pollFirst()
        val kvResidual = residuals(k)(vId)
        estimates(vId) += kvResidual
        residuals(k)(vId) = 0.0f
        val v = graph.getNodeById(vId).get
        val mass = heatKernelT * kvResidual / (k + 1.0f) / v.outboundCount
        for (uId <- v.outboundNodes()) {
          val u = graph.getNodeById(uId).get
          if (k + 1 == maxLength) {
            estimates(u.id) += kvResidual / v.outboundCount
          } else {
            val thresh = epsilon * u.outboundCount /
              (maxLength * phis(k)) / 2.0
            if (residuals(k+1)(uId) < thresh &&
              residuals(k+1)(uId) + mass >= thresh)
              largeResidualNodes.add((uId, k + 1))
            residuals(k+1)(uId) += mass
          }
        }
      }
    }
    estimates
  }

  def monteCarloHeatKernel(sourceId: Int, walkCount: Int): collection.Map[Int, Double] = {
    val visitProbabilities = CollectionsUtil.efficientIntDoubleMapWithDefault0()
    for (walkIndex <- 0 until walkCount) {
      val walk = samplePath(sourceId, maxLength)
      for (i <- 0 until maxLength) {
        visitProbabilities(walk(i)) += heatKernelProbability(i) / walkCount
      }
    }
    assert(math.abs(visitProbabilities.values.sum - 1.0) < 1.0e-3,
     "sum " + visitProbabilities.values.sum + " map " + visitProbabilities)
    visitProbabilities
  }
}
