package soal.ppr

import scala.collection.mutable
import com.twitter.cassovary.graph.{Node, DirectedGraph}
import java.util.Random
import soal.util.{ConstantDistribution, DiscreteDistribution}

/**
 * Contains methods related to personalized PageRank estimation.  All methods operate in the
 * context of the given graph, teleportProbability, and random.
 */

class BidirectionalPPRPrecomputor(val graph: DirectedGraph[Node],
                                 val teleportProbability: Float,
                                 val random: Random = new Random) {
  val estimator = new BidirectionalPPREstimator(graph, teleportProbability, random)

  /**
   * Estimates the personalized PageRank from the given source to the given target.  The result
   * will have mean relative error less than the given bound
   */
  def forwardPrecomputation(sourceId: Int,
                  minimumPPR: Float,
                  reverseThreshold: Float,
                  relativeError: Float = 0.1f,
                  guaranteeRelativeError: Boolean = false,
                  walkCountOption: Option[Int] = None): collection.Map[Int, Float] = {
    val chernoffConstant = if (guaranteeRelativeError)
      12 * math.exp(1) / math.log(2 / 1.0e-9) // guarantees the given relative error with
      // probability 1 - 1.0e-9
    else
      0.07 // Found in experiments to give mean relative error less than 10% when relativeError=10%

    val chernoffMultiplier = chernoffConstant / math.pow(relativeError, 2.0).toFloat
    val walkCount = walkCountOption.getOrElse(
      (chernoffMultiplier * reverseThreshold / minimumPPR).toInt)

    println(s"doing $walkCount walks")
    val walkDistribution = new mutable.HashMap[Int, Float]()
    for (walkIndex <- 0 until walkCount) {
      val v = estimator.samplePPR(sourceId)
      walkDistribution(v) = walkDistribution.getOrElse(v, 0.0f) + 1.0f / walkCount
    }
    walkDistribution
  }
}
