package soal.ppr

import com.twitter.cassovary.graph.{Node, DirectedGraph}
import scala.collection._
import soal.util.{ConstantDistribution, DiscreteDistribution, CollectionsUtil, DiscreteAliasSampler}
import java.util.Random

/** Contains methods for personalized PageRank search.
 */

class BidirectionalPPRSearcher(val graph: DirectedGraph[Node],
                             val teleportProbability: Float,
                             val random: Random = new Random) {
  val pprEstimator = new BidirectionalPPREstimator(graph, teleportProbability, random)

  /**
   * Computes and returns all data needed to quickly respond to searches on this target set.
   */
  def precomputeForTargetSet(targetIds: Seq[Int], maxResidual: Float
      ): TargetSetPrecomputation = {
    val reverseVectors = for (t <- targetIds) yield {
      val (estimates, residuals) = pprEstimator.computeContributions(t, maxResidual)
      // Combine estimates into residuals, using negative coordinates,
      // to create combined reverse vector
      for ((v, estimate) <- estimates) {
        residuals(-v - 1) = estimate
      }
      residuals
    }

    val reverseVectorSum = CollectionsUtil.sumMaps(reverseVectors)

    // Construct Samplers from each intermediate node v to the targets
    val targetSamplerMap = new mutable.HashMap[Int, DiscreteAliasSampler]()
    for (v <- reverseVectorSum.keysIterator
         if reverseVectorSum(v) > 1.0e-10f) {
      val targetIndicesFromV = (0 until targetIds.size) filter { i => reverseVectors(i)(v) > 0.0f }
      val targetsFromV = targetIndicesFromV map targetIds
      val valuesFromV = targetIndicesFromV map { i => reverseVectors(i)(v) }
      if (valuesFromV.sum < 1.0e-9f) {
        println(s"v=$v, totalResidual(v)=${reverseVectorSum(v)}")
      }
      targetSamplerMap(v) = new DiscreteAliasSampler(targetsFromV, valuesFromV, random)
    }

    new SimpleTargetSetPrecomputation(reverseVectorSum, targetSamplerMap)
  }

  /**
   * Given source and targetSetPrecomputations, returns a map from t to an estimate of the
   * personalized PageRank from source to t.
   */
  def sampleTargets(
      sourceDistribution: DiscreteDistribution,
      targetPrecomputations: TargetSetPrecomputation,
      walkCount: Int,
      sampleCount: Int
      ):  Map[Int, Float] = {
    // We will create a sampler over intermediate nodes v, where the sampling weight of v is
    // (an estimate of) personalizedPagerank(s, v) * reverseVectorSum(v)
    // and (to incorporate estimates) weight(-s - 1) = reverseVectorSum(-s - 1)
    // (generalizing  naturally if s is a distribution rather than a number).
    // .
    // Then to sample a target t we sample v from vWeights
    // followed by sampling t from targetPrecomputations.sampleTargetFrom(v).
    val vWeights = CollectionsUtil.efficientIntFloatMap().withDefaultValue(0.0f)
    for (walkIndex <- 0 until walkCount) {
      val v = pprEstimator.samplePPR(sourceDistribution)
      if (targetPrecomputations.reverseVectorSum(v) > 1.0e-9) {
        assert(targetPrecomputations.asInstanceOf[SimpleTargetSetPrecomputation].targetSamplerMap
          .contains(v),
          s"Error: node $v has totalResidual " +
            targetPrecomputations.reverseVectorSum(v) + " yet isn't in targetSamplerMap")
        vWeights(v) += 1.0f / walkCount * targetPrecomputations.reverseVectorSum(v)
      }
    }
    println(s"sampler completed $walkCount forward walks. Now sampling targets $sampleCount times.")

    // Note that "reverseVector" hear means a vector combining residual values with estimate values,
    // where the estimate values are stored in the negative coordinates
    for ((s, p) <- sourceDistribution.valuesAndProbabilities()) {
      val sEstimateIndex = -s - 1
      vWeights(sEstimateIndex) = targetPrecomputations.reverseVectorSum(sEstimateIndex)
    }

    val targetSampleCounts = CollectionsUtil.efficientIntFloatMapWithDefault0()

    if (vWeights.values.sum > 1.0e-9f) {
      val vWeightDistribution = DiscreteAliasSampler(vWeights, random)
      for (sampleIndex <- 0 until sampleCount) {
        val v = vWeightDistribution.sample()
        val t = targetPrecomputations.sampleTargetFrom(v)
        targetSampleCounts(t) += 1
      }
    } else { // vWeights has sum 0
      // In the edge case where none of the random walks reach nodes with positive residual,
      // the vWeights will have sum 0, and we return an empty map.
      println("Warning: vWeights had sum " + vWeights.values.sum + "; returning empty map.")
    }

    // Convert sample counts to actual PPR values
    val targetPPRSum = vWeights.values.sum
    // For some reason, just doing targetSampleCounts mapValues { _ / sampleCount * targetPPRSum}
    // doesn't preserve the default value.
    for (t <- targetSampleCounts.keys) {
      targetSampleCounts(t) = targetSampleCounts(t) / sampleCount * targetPPRSum
    }
    targetSampleCounts
  }

  /**
   * Convenience method for the case when source is a single node.
   */
  def sampleTargets(sourceId: Int,
                    targetPrecomputations: TargetSetPrecomputation,
                    walkCount: Int,
                    sampleCount: Int
                    ):  Map[Int, Float] =
    sampleTargets(new ConstantDistribution(sourceId), targetPrecomputations, walkCount, sampleCount)
}
