package soal.ppr.experiments

import java.util.Random

import com.twitter.cassovary.graph.{DirectedGraph, Node}
import net.openhft.koloboke.collect.map.hash.HashIntFloatMaps
import soal.ppr.BidirectionalPPREstimator

import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer

/*
 * Created by plofgren on 5/12/15.
 */

class PathLengthExperiments(graph: DirectedGraph[Node], teleportProb: Float,
                                    pageRanks: IndexedSeq[Double]) {
  val random = new Random()
  val pprEstimator = new BidirectionalPPREstimator(graph, teleportProb, random)

  def computeProbabilitiesByLength(s: Node, t: Node): mutable.IndexedSeq[Double] = {
    val probs = Array.fill(5)(0.0)
    if (s.outboundNodes contains(t.id)) {
      probs(1) = 1.0 / s.outboundCount
    }

    /*val sOutNeighbors = HashIntSets.newMutableSet()
    s.outboundNodes foreach (sOutNeighbors.add(_))
    for (vId <- t.inboundNodes) {
      if (sOutNeighbors.contains(vId)) {
        val vOutDegree = graph.getNodeById(vId).get.outboundCount
        probs(2) += 1.0 / s.outboundCount / vOutDegree
      }
    }
*/
    // sProb2(v) is the probability of being at v after a 2-step walk from s
    val sProb2 = HashIntFloatMaps.newMutableMap()
    for (uId <- s.outboundNodes) {
      val u = graph.getNodeById(uId).get
      for (vId <- u.outboundNodes) {
        sProb2.put(vId, sProb2.getOrDefault(vId, 0.0f) + 1.0f / s.outboundCount / u.outboundCount)
        //println(s"l2: ${s.id} -> $uId -> $vId weight 1.0 / ${s.outboundCount} / ${u.outboundCount}")
      }
    }
    probs(2) = sProb2.getOrDefault(t.id, 0.0f)

    for (zId <- t.inboundNodes) {
      val z = graph.getNodeById(zId).get

      if (sProb2.containsKey(zId)) {
        probs(3) += sProb2.get(zId) / z.outboundCount
        //println(s"l3: ${s.id} -> ? -> ${z.id} -> ${t.id} weight ${sProb2.get(zId)} / ${z.outboundCount}")
      }

      for (yId <- z.inboundNodes) {
        if (sProb2.containsKey(yId)) {
          val yOutDegree = graph.getNodeById(yId).get.outboundCount
          probs(4) +=  sProb2.get(yId) / yOutDegree / z.outboundCount
          //println(s"l4: ${s.id} -> ? -> $yId -> ${z.id} -> ${t.id} weight ${sProb2.get(yId)} / $yOutDegree / ${z.outboundCount}")
        }
      }
    }

    probs
  }

  def pathLengthProbExperiment(pairCount: Int): Unit = {
    val minimumPPR = 4.0f / graph.nodeCount
    val relativeError = 0.05f
    val starts = Array.fill(pairCount) {
      ExperimentUtils.randomNode(graph)
    }
    val targets = Array.fill(pairCount) {
      ExperimentUtils.sampleFromPageRank(graph, teleportProb)
    }

    println("   p0       p1       p2       p3       p4    l4+*pi(t) estimate truePPR relError")
    val results = (starts zip targets).par.map { case (s, t) =>
      val probs = computeProbabilitiesByLength(s, t)
      val result = new ArrayBuffer[Double]()
      for (i <- 0 until probs.size) {
        result += probs(i) * teleportProb * math.pow(1.0 - teleportProb, i)
      }
      result += math.pow(1.0 - teleportProb, probs.size) * pageRanks(t.id)
      val estimate = result.sum
      val truePPR = pprEstimator.estimatePPR(s.id, t.id, minimumPPR, relativeError)
      result += estimate
      result += truePPR
      result += (estimate - truePPR) / truePPR
      assert(result.size == 9)
      println((result map ("%8g".format(_))).mkString(" "))
      result
    }.seq
    println("allResults = " + ExperimentUtils.matrixToPythonList(results))
    val highPPRResults = results filter (_(7) > minimumPPR)
    println("highPPRResults = " + ExperimentUtils.matrixToPythonList(highPPRResults))
    println("relErrors = " + (highPPRResults map (_(8))))
    println("meanAbsRelError = " + ExperimentUtils.mean(highPPRResults map {r => math.abs(r(8))}))
  }
}

object PathLengthExperiments {
  def main(args: Array[String]): Unit = {
    val graph = ExperimentUtils.readGraph("src/test/resources/test_graph.txt")
    val s = graph.getNodeById(6).get
    val t = graph.getNodeById(1).get
    val probs = new PathLengthExperiments(graph, 0.2f, IndexedSeq()).computeProbabilitiesByLength(s, t)
    println("probs: " + probs)
    assert(probs.equals(ArrayBuffer(0, 1.0 / 2, 0.0, 7.0 / 16, 1.0 / 8)))
  }
}