package soal.ppr.experiments

import com.twitter.cassovary.graph.{DirectedGraph, Node}
import soal.ppr.{BidirectionalPPREstimator, BidirectionalUndirectedPPREstimator}
import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer

class UndirectedPPRExperiments(graph: DirectedGraph[Node], teleportProb: Float) {
  val directedEstimator = new BidirectionalPPREstimator(graph, teleportProb)
  val undirectedEstimator = new BidirectionalUndirectedPPREstimator(graph, teleportProb)
  val relErrorForTruth = 0.01f
  val minimumProbability = 1.0f / graph.nodeCount

  def sampleValidPairAndTruePPR(): (Node, Node, Float) = {
    val start = ExperimentUtils.randomNode(graph)
    val target = ExperimentUtils.sampleFromPageRank(graph, teleportProb)
    val truePPR = directedEstimator.estimatePPR(start.id, target.id,
      minimumProbability, relErrorForTruth)
    if (truePPR > minimumProbability) {
      (start, target, truePPR)
    } else {
      println("rejecting target with ppr/minPPR = " + truePPR / minimumProbability)
      sampleValidPairAndTruePPR()
    }
  }

  def accuracyExperiment(trialCount: Int): Unit = {
    val relErrorParams = Seq(0.2f, 0.1f, 0.05f)

    val meanRelativeErrors = new ArrayBuffer[Double]()

    for (relErrorParam <- relErrorParams) {
      val relativeErrors = new ArrayBuffer[Double]()
      val trueValues = new ArrayBuffer[Double]()
      val estimatedValues = new ArrayBuffer[Double]()

      (0 until trialCount) map { trialIndex =>
        val (s, t, trueValue) = sampleValidPairAndTruePPR()
        val estimate = undirectedEstimator.estimatePPR(s.id, t.id, minimumProbability,
          relErrorParam)
        val signedRelError = (estimate - trueValue) / trueValue
        println(s"true\tbi\trelError: " +
          s"$trueValue\t$estimate\t$signedRelError")
        trueValues += trueValue
        relativeErrors += math.abs(signedRelError)
      }
      println(s"\nrelErrorTarget=$relErrorParam")

      println("meanRelError: " +
      ExperimentUtils.mean(relativeErrors))
      meanRelativeErrors += ExperimentUtils.mean(relativeErrors)
    }
    println("relErrorTargets: " + relErrorParams)
    println("meanRelErrors: " + meanRelativeErrors)
  }

  def walkPushTimesTimesExperiment(trialCount: Int, msPerWalks: Seq[Float]): Unit = {
    val walkPushTimes = msPerWalks map {measureMeanWalkPushTime(trialCount, _)}
    println("msPerWalks " + ExperimentUtils.toPythonList(msPerWalks))
    println("walkTimes " + ExperimentUtils.toPythonList(walkPushTimes map {_._1}))
    println("pushTimes " + ExperimentUtils.toPythonList(walkPushTimes map {_._2}))
  }

  def measureMeanWalkPushTime(trialCount: Int, msPerWalk: Float): (Double, Double) = {

    val walkTimes = mutable.ArrayBuffer[Float]()
    val pushTimes = mutable.ArrayBuffer[Float]()
    for (trialIndex <- 0 until trialCount) {
      val start = ExperimentUtils.randomNode(graph)
      val target = ExperimentUtils.sampleFromPageRank(graph, teleportProb)
      val (walkTime, pushTime) = measureForwardpushTime(start, target, msPerWalk)
      walkTimes.append(walkTime)
      pushTimes.append(pushTime)
    }

    val totalTimes = (walkTimes zip pushTimes) map { case (x, y) => x + y }
    val meanwalkTime = ExperimentUtils.mean(walkTimes)
    val meanpushTime = ExperimentUtils.mean(pushTimes)
    println("pushTimes = " + pushTimes.mkString("[", ", ", "]"))
    println("walkTimes = " + walkTimes.mkString("[", ", ", "]"))
    println("totalTimes = " + totalTimes.mkString("[", ", ", "]"))
    println("meanwalkTime = " + meanwalkTime)
    println("meanpushTime = " + meanpushTime)
    println("math.sqrt(meanpushTime / meanwalkTime): " + math.sqrt(meanpushTime / meanwalkTime))
    (meanwalkTime, meanpushTime)
  }

  def measureForwardpushTime(start: Node,
                                target: Node,
                                msPerWalk: Float): (Float, Float) = {
    // This is copied and pasted. TODO: find a way of measure forward+backward time without copying+pasting
    val minimumPPR = 4.0f / graph.nodeCount
    val chernoffMultiplier = 7.0f // TODO!

    def computeWalkCount(maxResidualRatio: Float): Int =
      (chernoffMultiplier * target.outboundCount * maxResidualRatio / minimumPPR).toInt
    def estimatewalkTimeInMillis(maxResidual: Float): Float =
      computeWalkCount(maxResidual) * msPerWalk

    val ((estimates, residualRatios, maxResidualRatio), pushTime) = ExperimentUtils.time {
      undirectedEstimator.balancedForwardResiduals(start.id, estimatewalkTimeInMillis)
    }
    
    val walkCount = computeWalkCount(maxResidualRatio)
    var estimate = estimates(target.id)
    val (_, walkTime) = ExperimentUtils.time {
      for (walkIndex <- 0 until walkCount) {
        val v = undirectedEstimator.samplePPR(target.id)
        estimate += target.outboundCount * residualRatios(v) / walkCount
      }
    }

    println(s"walkCount forwardResidualSize walkTime pushTime " +
      s"$walkCount ${residualRatios.size + estimates.size} $walkTime $pushTime" )
    (walkTime.toFloat, pushTime.toFloat)
  }
/*
  def measureAccuracyOnGraphs(graphNames: Seq[String], trialCount: Int): Unit = {
    val accuracyTriples = new ArrayBuffer[(Double, Double, Double)]()
    for (graphName <- graphNames) yield {
      println("graph " + graphName)
      val graph = ExperimentUtils.readGraph(graphName)
      accuracyTriples += accuracyExperiment(graph, trialCount)
      val (times1, times2, times3) = accuracyTriples.unzip3
      println("------accuracy summary-------")
      for (accuracies <- List(times1, times2, times3)) {
        println(accuracies.mkString("\t"))
      }
      for (accuracies <- List(times1, times2, times3)) {
        println(ExperimentUtils.toPythonList(accuracies) + ",")
      }
    }
  }

  def meanRelativeError(trueValues: Seq[Double], estimatedValues: Seq[Double]): Double = {
    relativeErrors(trueValues, estimatedValues).sum / trueValues.size
  }

  def relativeErrors(trueValues: Seq[Double], estimatedValues: Seq[Double]): Seq[Double] = {
    (trueValues zip estimatedValues) map { case (t, e) =>
      math.abs(t - e) / t
    }
  }
  */
}
